@@ -52,7 +52,9 @@ extern "C"
 #include <time.h>
 #include <signal.h>
 #include <getopt.h>
+#include <stdarg.h>
 
+#define __APPLE_USE_RFC_3542 1
 #include <netinet/in.h>
 
 #include "config.h"
@@ -132,7 +134,7 @@ typedef struct ping_data
 #define MAX_PING_DATA   ( MAX_IP_PACKET - SIZE_IP_HDR - SIZE_ICMP_HDR )
 
 /* sized so as to be like traditional ping */
-#define DEFAULT_PING_DATA_SIZE  ( MIN_PING_DATA + 44 ) 
+#define DEFAULT_PING_DATA_SIZE  56
 
 /* maxima and minima */
 #define MAX_COUNT               10000
@@ -312,6 +314,7 @@ int verbose_flag, quiet_flag, stats_flag, unreachable_flag, alive_flag;
 int elapsed_flag, version_flag, count_flag, loop_flag;
 int per_recv_flag, report_all_rtts_flag, name_flag, addr_flag, backoff_flag;
 int multif_flag;
+int timestamp_flag = 0;
 #if defined( DEBUG ) || defined( _DEBUG )
 int randomly_lose_flag, sent_times_flag, trace_flag, print_per_system_flag;
 int lose_factor;
@@ -328,7 +331,6 @@ void add_addr( char *name, char *host, struct in_addr ipaddr );
 void add_addr( char *name, char *host, FPING_SOCKADDR *ipaddr );
 #endif
 char *na_cat( char *name, struct in_addr ipaddr );
-char *cpystr( char *string );
 void crash_and_burn( char *message );
 void errno_crash_and_burn( char *message );
 char *get_host_by_address( struct in_addr in );
@@ -353,6 +355,7 @@ HOST_ENTRY *ev_dequeue();
 void ev_remove(HOST_ENTRY *h);
 void add_cidr(char *);
 void add_range(char *, char *);
+void print_warning(char *fmt, ...);
 
 /*** function definitions ***/
 
@@ -490,7 +493,7 @@ int main( int argc, char **argv )
 
     /* get command line options */
 
-    while( ( c = getopt( argc, argv, "gedhlmnqusaAvz:t:H:i:p:f:r:c:b:C:Q:B:S:I:T:O:" ) ) != EOF )
+    while( ( c = getopt( argc, argv, "gedhlmnqusaAvDz:t:H:i:p:f:r:c:b:C:Q:B:S:I:T:O:" ) ) != EOF )
     {
         switch( c )
         {
@@ -581,6 +584,10 @@ int main( int argc, char **argv )
             stats_flag = 1;
             break;
 
+        case 'D':
+            timestamp_flag = 1;
+            break;
+
         case 'l':
             loop_flag = 1;
             backoff_flag = 0;
@@ -682,7 +689,7 @@ int main( int argc, char **argv )
 
     /* validate various option settings */
 
-    if (ttl < 0 || ttl > 255) {  
+    if (ttl > 255) {
         fprintf(stderr, "ping: ttl %u out of range\n", ttl);  
         usage(1);
     }  
@@ -858,7 +865,6 @@ int main( int argc, char **argv )
         FILE *ping_file;
         char line[132];
         char host[132];
-        char *p;
         
         if( strcmp( filename, "-" ) == 0 )
             ping_file = fdopen( 0, "r" );
@@ -877,9 +883,7 @@ int main( int argc, char **argv )
             if( ( !*host ) || ( host[0] == '#' ) )  /* magic to avoid comments */
                 continue;
             
-            p = cpystr( host );
-            add_name( p );
-        
+            add_name(host);
         }/* WHILE */
         
         fclose( ping_file );
@@ -1030,7 +1034,7 @@ void add_cidr(char *addr)
         char buffer[20];
         in_addr_tmp.s_addr = htonl(net_addr);
         inet_ntop(AF_INET, &in_addr_tmp, buffer, sizeof(buffer));
-        add_name(cpystr(buffer));
+        add_name(buffer);
     }
 
     freeaddrinfo(addr_res);
@@ -1080,7 +1084,7 @@ void add_range(char *start, char *end)
         char buffer[20];
         in_addr_tmp.s_addr = htonl(start_long);
         inet_ntop(AF_INET, &in_addr_tmp, buffer, sizeof(buffer));
-        add_name(cpystr(buffer));
+        add_name(buffer);
         start_long++;
     }
 }
@@ -1144,7 +1148,7 @@ void main_loop()
                     ev_enqueue(h);
                 }
                 /* Count mode: schedule timeout after last ping */
-                else if(count_flag && count_flag && h->num_sent >= count) {
+                else if(count_flag && h->num_sent >= count) {
                     h->ev_type = EV_TYPE_TIMEOUT;
                     h->ev_time.tv_sec = last_send_time.tv_sec;
                     h->ev_time.tv_usec = last_send_time.tv_usec;
@@ -1575,7 +1579,12 @@ int send_ping( int s, HOST_ENTRY *h )
         ( struct sockaddr* )&h->saddr, sizeof( FPING_SOCKADDR ) );
 
     if( n < 0 || n != ping_pkt_size )
-    {
+    if(
+        (n < 0 || n != ping_pkt_size)
+#if defined( EHOSTDOWN )
+        && errno != EHOSTDOWN
+#endif
+    ) {
         if( verbose_flag || unreachable_flag )
         {
             printf( "%s", h->host );
@@ -1586,10 +1595,11 @@ int send_ping( int s, HOST_ENTRY *h )
 
         }/* IF */
         
-        num_unreachable++;
-        remove_job( h ); 
+        h->num_sent++;
+        h->num_sent_i++;
+        num_pingsent++;
         free( buffer );
-        return(0);
+        return(1);
     }
 
     /* mark this trial as outstanding */
@@ -1823,6 +1833,11 @@ int wait_for_reply(long wait_time)
 
     if( per_recv_flag )
     {
+        if(timestamp_flag) {
+            printf("[%lu.%06lu] ",
+                 (unsigned long)current_time.tv_sec,
+                 (unsigned long)current_time.tv_usec);
+        }
         avg = h->total_time / h->num_recv;
         printf( "%s%s : [%d], %d bytes, %s ms",
             h->host, h->pad, this_count, result, sprint_tm( this_reply ) );
@@ -1915,7 +1930,7 @@ int handle_random_icmp( FPING_ICMPHDR *p, int psize, FPING_SOCKADDR *addr )
             
             if( p->icmp_code > ICMP_UNREACH_MAXTYPE )
             {
-                fprintf( stderr, "ICMP Unreachable (Invalid Code) from %s for ICMP Echo sent to %s",
+                print_warning("ICMP Unreachable (Invalid Code) from %s for ICMP Echo sent to %s",
                     inet_ntoa( addr->sin_addr ), h->host );
 
 #else
@@ -1928,13 +1943,13 @@ int handle_random_icmp( FPING_ICMPHDR *p, int psize, FPING_SOCKADDR *addr )
             
             if( p->icmp6_code > ICMP_UNREACH_MAXTYPE )
             {
-                fprintf( stderr, "ICMP Unreachable (Invalid Code) from %s for ICMP Echo sent to %s",
+                print_warning("ICMP Unreachable (Invalid Code) from %s for ICMP Echo sent to %s",
                     addr_ascii, h->host );
 #endif
             }/* IF */
             else
             {
-                fprintf( stderr, "%s from %s for ICMP Echo sent to %s",
+                print_warning("%s from %s for ICMP Echo sent to %s",
 #ifndef IPV6
                     icmp_unreach_str[p->icmp_code], inet_ntoa( addr->sin_addr ), h->host );
 #else
@@ -1945,12 +1960,12 @@ int handle_random_icmp( FPING_ICMPHDR *p, int psize, FPING_SOCKADDR *addr )
 
             if( inet_addr( h->host ) == -1 )
 #ifndef IPV6
-                fprintf( stderr, " (%s)", inet_ntoa( h->saddr.sin_addr ) );
+                print_warning(" (%s)", inet_ntoa( h->saddr.sin_addr ) );
 #else
-                fprintf( stderr, " (%s)", addr_ascii);
+                print_warning(" (%s)", addr_ascii);
 #endif
             
-            fprintf( stderr, "\n" );
+            print_warning("\n" );
         
         }/* IF */
 
@@ -2088,11 +2103,14 @@ void add_name( char *name )
         /* input name is an IP addr, go with it */
         if( name_flag )
         {
-            if( addr_flag )
-                add_addr( name, na_cat( get_host_by_address( *ipa ), *ipa ), *ipa );
+            if( addr_flag ) {
+                char namebuf[256];
+                snprintf(namebuf, 256, "%s (%s)", get_host_by_address(*ipa), name);
+                add_addr(name, namebuf, *ipa);
+            }
             else
             {
-                nm = cpystr( get_host_by_address( *ipa ) );
+                nm = get_host_by_address( *ipa );
                 add_addr( name, nm, *ipa );
 
             }/* ELSE */
@@ -2125,36 +2143,39 @@ void add_name( char *name )
             if( getnetgrent( &machine, &user_ignored, &domain_ignored ) == 0 )
             {
                 endnetgrent();
-                if( !quiet_flag )
-                    fprintf( stderr, "%s address not found\n", name );
+                print_warning("%s address not found\n", name );
                 
                 num_noaddress++;
                 return;
             
             }/* IF */
             else
-                add_name( cpystr( machine ) );
+                add_name(machine);
 
             while( getnetgrent( &machine, &user_ignored, &domain_ignored ) )
-                add_name( cpystr( machine ) );
+                add_name(machine);
       
             endnetgrent();
             return;
 #else
-            if( !quiet_flag )
-                fprintf( stderr, "%s address not found\n", name );
+            print_warning("%s address not found\n", name );
             
             num_noaddress++;
             return ; 
 #endif /* NIS_GROUPS */
         }/* IF */
     }/* IF */
+
+    if(host_ent->h_addrtype != AF_INET) {
+        print_warning("%s: IPv6 address returned by gethostbyname (options inet6 in resolv.conf?)\n", name );
+        num_noaddress++;
+        return; 
+    }
   
     host_add = ( struct in_addr* )*( host_ent->h_addr_list ); 
     if( host_add == NULL )
     { 
-        if( !quiet_flag )
-            fprintf( stderr, "%s has no address data\n", name );
+        print_warning("%s has no address data\n", name );
         
         num_noaddress++;
         return; 
@@ -2165,11 +2186,15 @@ void add_name( char *name )
         /* it is indeed a hostname with a real address */
         while( host_add )
         {
-            if( name_flag && addr_flag )
-                add_addr( name, na_cat( name, *host_add ), *host_add );
+            if( name_flag && addr_flag ) {
+                char namebuf[256];
+                nm = inet_ntoa(*host_add);
+                snprintf(namebuf, 256, "%s (%s)", name, nm);
+                add_addr( name, namebuf, *host_add );
+            }
             else if( addr_flag )
             {
-                nm = cpystr( inet_ntoa( *host_add ) );
+                nm = inet_ntoa(*host_add);
                 add_addr( name, nm, *host_add );
             }/* ELSE IF */
             else
@@ -2183,76 +2208,78 @@ void add_name( char *name )
         }/* WHILE */
     }/* ELSE */
 #else
-    FPING_SOCKADDR  dst;
-    struct addrinfo     *res, hints;
-    int                     ret_ga;
-    char                        *hostname;
-    size_t len;
+    FPING_SOCKADDR    dst;
+    struct addrinfo   *res, hints;
+    int               ret_ga;
+    size_t            len;
+    char              *printname;
+    char              namebuf[256];
+    char              addrbuf[256];
 
     /* getaddrinfo */
     bzero(&hints, sizeof(struct addrinfo));
-    hints.ai_flags = name_flag ? AI_CANONNAME : 0;
+    hints.ai_flags = 0;
     hints.ai_family = AF_INET6;
     hints.ai_socktype = SOCK_RAW;
     hints.ai_protocol = IPPROTO_ICMPV6;
-
     ret_ga = getaddrinfo(name, NULL, &hints, &res);
     if (ret_ga) {
         if(!quiet_flag)
-            warnx("%s", gai_strerror(ret_ga));
-        num_noaddress++;
-        return; 
-    }
-    if (res->ai_canonname) hostname = res->ai_canonname;
-    else hostname = name;
-    if (!res->ai_addr) {
-        if(!quiet_flag)
-            warnx("getaddrinfo failed");
+            print_warning("%s: %s\n", name, gai_strerror(ret_ga));
         num_noaddress++;
         return; 
     }
     len = res->ai_addrlen;
     if (len > sizeof(FPING_SOCKADDR)) len = sizeof(FPING_SOCKADDR);
     (void)memcpy(&dst, res->ai_addr, len);
-    add_addr(name, name, &dst);
+
+    /* name_flag: addr -> name lookup requested) */
+    if(!name_flag) {
+        printname = name;
+    }
+    else {
+        int ret;
+        ret = getnameinfo(res->ai_addr, res->ai_addrlen, namebuf,
+                          sizeof(namebuf)/sizeof(char), NULL, 0, 0);
+        if (ret) {
+            if(!quiet_flag) {
+                print_warning("%s: %s\n", name, gai_strerror(ret_ga));
+            }
+            num_noaddress++;
+            return; 
+        }
+        printname = namebuf;
+    }
+
+    /* addr_flag: name -> addr lookup requested */
+    if(addr_flag) {
+        int ret;
+        ret = getnameinfo(res->ai_addr, res->ai_addrlen, addrbuf,
+                          sizeof(addrbuf)/sizeof(char), NULL, 0, NI_NUMERICHOST);
+        if (ret) {
+            if(!quiet_flag) {
+                print_warning("%s: %s\n", name, gai_strerror(ret_ga));
+            }
+            num_noaddress++;
+            return; 
+        }
+
+        if(name_flag) {
+            char nameaddrbuf[512];
+            snprintf(nameaddrbuf, sizeof(nameaddrbuf)/sizeof(char), "%s (%s)", printname, addrbuf);
+            add_addr(name, nameaddrbuf, &dst);
+        }
+        else {
+            add_addr(name, addrbuf, &dst);
+        }
+    }
+    else {
+        add_addr(name, printname, &dst);
+    }
 #endif
 } /* add_name() */
 
 
-/************************************************************
-
-  Function: na_cat
-
-*************************************************************
-
-  Inputs:  char* name, struct in_addr ipaddr
-
-  Returns:  char*
-
-  Description:
-
-************************************************************/
-
-char *na_cat( char *name, struct in_addr ipaddr )
-{
-    char *nm, *as;
-
-    as = inet_ntoa( ipaddr );
-    nm = ( char* )malloc( strlen( name ) + strlen( as ) + 4 );
-
-    if( !nm )
-        crash_and_burn( "can't allocate some space for a string" );
-    
-    strcpy( nm, name );
-    strcat( nm, " (" );
-    strcat( nm, as );
-    strcat( nm, ")" );
-
-    return( nm );
-
-} /* na_cat() */
-
-
 /************************************************************
 
   Function: add_addr
@@ -2283,8 +2310,8 @@ void add_addr( char *name, char *host, FPING_SOCKADDR *ipaddr )
 
     memset( ( char* ) p, 0, sizeof( HOST_ENTRY ) );
 
-    p->name = name;
-    p->host = host;
+    p->name = strdup(name);
+    p->host = strdup(host);
 #ifndef IPV6
     p->saddr.sin_family = AF_INET;
     p->saddr.sin_addr = ipaddr; 
@@ -2398,41 +2425,6 @@ char *get_host_by_address( struct in_addr in )
 
 } /* get_host_by_address() */
 
-
-/************************************************************
-
-  Function: cpystr
-
-*************************************************************
-
-  Inputs:  char* string
-
-  Returns:  char*
-
-  Description:
-
-************************************************************/
-
-char *cpystr( char *string )
-{
-    char *dst;
-
-    if( string )
-    {
-        dst = ( char* )malloc( 1 + strlen( string ) );
-        if( !dst )
-            crash_and_burn( "can't allocate some space for a string" );
-        
-        strcpy( dst, string );
-        return dst;
-    
-    }/* IF */
-    else 
-        return NULL;
-
-} /* cpystr() */
-
-
 /************************************************************
 
   Function: crash_and_burn
@@ -2474,6 +2466,25 @@ void errno_crash_and_burn( char *message )
 } /* errno_crash_and_burn() */
 
 
+/************************************************************
+
+  Function: print_warning
+
+  Description: fprintf(stderr, ...), unless running with -q
+
+*************************************************************/
+
+void print_warning(char *format, ...) {
+    va_list args;
+    if(!quiet_flag) {
+        va_start(args, format );
+        vfprintf(stderr, format, args);
+        va_end(args);
+    }
+}
+
+
+
 /************************************************************
 
   Function: timeval_diff
@@ -2539,36 +2550,32 @@ char * sprint_tm( int t )
 {
     static char buf[10];
 
-    /* <= 0.99 ms */
-    if( t < 100 )
-    {
+    if( t < 0 ) {
+        /* negative (unexpected) */
+        sprintf( buf, "%.2g", (double) t / 100 );
+    }
+    else if( t < 100 ) {
+        /* <= 0.99 ms */
         sprintf( buf, "0.%02d", t );
-        return( buf );
-
-    }/* IF */
-
-    /* 1.00 - 9.99 ms */
-    if( t < 1000 )
-    {
+    }
+    else if( t < 1000 ) {
+        /* 1.00 - 9.99 ms */
         sprintf( buf, "%d.%02d", t / 100, t % 100 );
-        return( buf );
-
-    }/* IF */
-
-    /* 10.0 - 99.9 ms */
-    if( t < 10000 )
-    {
+    }
+    else if( t < 10000 ) {
+        /* 10.0 - 99.9 ms */
         sprintf( buf, "%d.%d", t / 100, ( t % 100 ) / 10 );
-        return( buf );
-    
-    }/* IF */
-  
-    /* >= 100 ms */
-    sprintf( buf, "%d", t / 100 );
+    }
+    else if( t < 100000000 ) {
+        /* 100 - 1'000'000 ms */
+        sprintf( buf, "%d", t / 100 );
+    }
+    else {
+        sprintf( buf, "%.2e", (double) (t / 100) );
+    }
+
     return( buf );
-
-} /* sprint_tm() */
-
+}
 
 /************************************************************
 
@@ -2810,10 +2817,11 @@ void usage(int is_error)
     fprintf(out, "Usage: %s [options] [targets...]\n", prog );
     fprintf(out, "   -a         show targets that are alive\n" );
     fprintf(out, "   -A         show targets by address\n" );
-    fprintf(out, "   -b n       amount of ping data to send, in bytes (default %d)\n", ping_data_size );
+    fprintf(out, "   -b n       amount of ping data to send, in bytes (default %d)\n", DEFAULT_PING_DATA_SIZE);
     fprintf(out, "   -B f       set exponential backoff factor to f\n" );
     fprintf(out, "   -c n       count of pings to send to each target (default %d)\n", count );  
     fprintf(out, "   -C n       same as -c, report results in verbose format\n" );
+    fprintf(out, "   -D         print timestamp before each output line\n" );
     fprintf(out, "   -e         show elapsed time on return packets\n" );
     fprintf(out, "   -f file    read list of targets from a file ( - means stdin) (only if no -g specified)\n" );
     fprintf(out, "   -g         generate target list (only if no -f specified)\n" );
